import logging

import boto3
from instance_control import config
from instance_control import volume

from instance_control.aws import ec2_node
from instance_control.command import Command

_LOG = logging.getLogger('bubuku.cluster.command.attach')


class AttachCommand(Command):
    def __init__(self, cluster_config_path: str, volume_id: str):
        super().__init__(cluster_config_path)
        self.volume_id = volume_id

    def execute(self):
        ec2_client = boto3.client('ec2', region_name=self.cluster_config['region'])
        ec2_resource = boto3.resource('ec2', region_name=self.cluster_config['region'])
        vol = ec2_resource.Volume(self.volume_id)
        volume.check_volume_available(vol)
        vol.create_tags(Tags=[{'Key': 'Name', 'Value': config.KAFKA_LOGS_EBS}])
        _LOG.info('Created tag:Name=%s for %s', config.KAFKA_LOGS_EBS, self.volume_id)

        self.cluster_config['create_ebs'] = False
        self.cluster_config['availability_zone'] = vol.availability_zone
        self.cluster_config['cluster_size'] = 1

        ec2_node.create(self.cluster_config)
        volume.wait_volumes_attached(ec2_client, ec2_resource)



import logging

import boto3

_LOG = logging.getLogger('bubuku.cluster.aws.metric')


def create_auto_recovery_alarm(cluster_config: dict, instance_id):
    _LOG.info('Creating AWS auto recovery alarm for %s', instance_id)
    cw = boto3.client('cloudwatch', region_name=cluster_config['region'])
    alarm_actions = ['arn:aws:automate:{}:ec2:recover'.format(cluster_config['region'])]

    alarm_name = '{}-{}-auto-recover'.format(cluster_config['cluster_name'], instance_id)
    cw.put_metric_alarm(AlarmName=alarm_name,
                        AlarmActions=alarm_actions,
                        MetricName='StatusCheckFailed_System',
                        Namespace='AWS/EC2',
                        Statistic='Minimum',
                        Dimensions=[{
                            'Name': 'InstanceId',
                            'Value': instance_id
                        }],
                        Period=60,  # 1 minute
                        EvaluationPeriods=2,
                        Threshold=0,
                        ComparisonOperator='GreaterThanThreshold')
    _LOG.info('Created alarm %s', alarm_name)



#!/usr/bin/env python3

import copy
import logging
import time

import boto3
import yaml

from instance_control import config
from instance_control.aws import security_group, iam, subnet, metric
from instance_control.aws import taupage

_LOG = logging.getLogger('bubuku.cluster.aws.ec2_node')


def _create_tagged_volume(ec2_client: object, cluster_config: dict, zone: str, name: str):
    _LOG.info('Creating EBS volume %s in %s', name, zone)
    ebs_data = {
        "AvailabilityZone": zone,
        "VolumeType": cluster_config['volume_type'],
        "Size": cluster_config['volume_size'],
        "Encrypted": False, }
    vol = ec2_client.create_volume(**ebs_data)
    _LOG.info('%s is successfully created', vol['VolumeId'])

    _LOG.info('Tagging %s with Taupage:erase-on-boot, to format only once', vol['VolumeId'])
    tags = [{'Key': 'Name',
             'Value': name},
            {'Key': 'Taupage:erase-on-boot',
             'Value': 'True'}]
    ec2_client.create_tags(Resources=[vol['VolumeId']], Tags=tags)


def _launch_instance(region: str, ip: dict, ami: object, subnet: dict, security_group_id: str, cluster_config: dict):
    _LOG.info('Launching node %s in %s', ip['_defaultIp'], subnet['AvailabilityZone'])
    ec2_client = boto3.client('ec2', region_name=region)

    #
    # Override any ephemeral volumes with NoDevice mapping,
    # otherwise auto-recovery alarm cannot be actually enabled.
    #
    _LOG.info('Overriding ephemeral volumes to be able to set up AWS auto recovery alarm ')
    block_devices = []
    for bd in ami.block_device_mappings:
        if 'Ebs' in bd:
            #
            # This has to be our root EBS.
            #
            # If the Encrypted flag is present, we have to delete
            # it even if it matches the actual snapshot setting,
            # otherwise amazon will complain rather loudly.
            #
            # Take a deep copy before deleting the key:
            #
            bd = copy.deepcopy(bd)

            root_ebs = bd['Ebs']
            if 'Encrypted' in root_ebs:
                del (root_ebs['Encrypted'])

            block_devices.append(bd)
        else:
            # ignore any ephemeral volumes (aka. instance storage)
            block_devices.append({'DeviceName': bd['DeviceName'],
                                  'NoDevice': ''})

    if cluster_config['create_ebs']:
        _create_tagged_volume(ec2_client, cluster_config, subnet['AvailabilityZone'], config.KAFKA_LOGS_EBS)

    user_data = cluster_config['user_data']
    user_data['volumes']['ebs']['/dev/xvdk'] = config.KAFKA_LOGS_EBS
    taupage_user_data = '#taupage-ami-config\n{}'.format(yaml.safe_dump(user_data))

    resp = ec2_client.run_instances(
        ImageId=ami.id,
        MinCount=1,
        MaxCount=1,
        SecurityGroupIds=[security_group_id],
        UserData=taupage_user_data,
        InstanceType=cluster_config['instance_type'],
        SubnetId=subnet['SubnetId'],
        PrivateIpAddress=ip['PrivateIp'],
        BlockDeviceMappings=block_devices,
        IamInstanceProfile={'Arn': cluster_config['instance_profile']['Arn']},
        DisableApiTermination=False,
        EbsOptimized=True)

    instance = resp['Instances'][0]
    instance_id = instance['InstanceId']
    _LOG.info('Instance %s launched, waiting for it to initialize', instance_id)

    ec2_client.create_tags(Resources=[instance_id],
                           Tags=[{'Key': 'Name', 'Value': cluster_config['cluster_name']}])

    # wait for instance to initialize before we can assign a
    # public IP address to it or tag the attached volume
    while True:
        resp = ec2_client.describe_instances(InstanceIds=[instance_id])
        instance = resp['Reservations'][0]['Instances'][0]
        if instance['State']['Name'] != 'pending':
            _LOG.info('Instance %s is up', instance_id)
            break
        _LOG.info('Waiting 5 secs more ...')
        time.sleep(5)

    metric.create_auto_recovery_alarm(cluster_config, instance_id)


def _launch_nodes(cluster_config: dict):
    i = 0
    for ip in cluster_config['node_ips']:
        subnets = cluster_config['subnets']
        _launch_instance(cluster_config['region'], ip,
                         ami=cluster_config['taupage_amis'],
                         subnet=subnets[i % len(subnets)],
                         security_group_id=cluster_config['security_group']['GroupId'],
                         cluster_config=cluster_config)
        if i + 1 == cluster_config['cluster_size']:
            _LOG.info("Creation is finished")
            break
        i += 1
        _LOG.info("Sleeping for one minute before launching next node..")
        time.sleep(60)


def create(cluster_config: dict):
    artifact_name = 'bubuku-appliance'
    cluster_config['docker_image'] = 'registry.opensource.zalan.do/aruha/{}:{}'.format(artifact_name,
                                                                                       cluster_config['image_version'])
    _LOG.info('Preparing AWS configuration for ec2 instance creation')
    try:
        cluster_config['taupage_amis'] = taupage.find_amis(cluster_config['region'])
        cluster_config['subnets'] = subnet.get_subnets('internal-', cluster_config)
        cluster_config['node_ips'] = subnet.allocate_ip_addresses(cluster_config)
        cluster_config['security_group'] = security_group.create_or_ger_security_group(cluster_config)
        cluster_config['user_data'] = taupage.generate_user_data(cluster_config)
        cluster_config['instance_profile'] = iam.create_or_get_instance_profile(cluster_config)

        _launch_nodes(cluster_config)

    except:
        _LOG.error('''
            You were trying to deploy Bubuku, but the process has failed :-(
            
            One of the reasons might be that some of Private IP addresses we were
            going to use to launch the EC2 instances were taken by some other
            instances in the middle of the process.  If that is the case, simply
            retrying the operation might resolve the problem (you still might need
            to clean up after this attempt before retrying).
            
            Please review the error message to see if that is the case, then
            either correct the error or retry.
            
        ''')
        raise



import json
import logging
import subprocess

from bubuku.config import KafkaProperties
from bubuku.id_generator import BrokerIdGenerator
from bubuku.zookeeper import BukuExhibitor

_LOG = logging.getLogger('bubuku.broker')


class LeaderElectionInProgress(Exception):
    pass


class KafkaProcessHolder(object):
    def __init__(self):
        self.process = None

    def get(self):
        return self.process

    def set(self, process):
        self.process = process


class StartupTimeout(object):
    def __init__(self, initial_value: float, config: str):
        self.timeout = initial_value
        self.config = config

    def get_timeout(self) -> float:
        return self.timeout

    def on_timeout_fail(self):
        self.timeout += self.get_step()

    def get_step(self) -> float:
        return 0.

    def __str__(self):
        return 'timeout={}, step={}, config={}'.format(self.get_timeout(), self.get_step(), self.config)

    @staticmethod
    def build(props: dict):
        type_ = props.get('type', 'linear')
        result = StartupTimeout(float(props.get('initial', '300')),
                                ','.join('{}={}'.format(k, v) for k, v in props.items()))
        if type_ == 'linear':
            step = float(props.get('step', '60'))
            result.get_step = lambda: step
        elif type_ == 'progressive':
            scale = float(props.get('step', '0.5'))
            result.get_step = lambda: result.timeout * scale
        else:
            raise NotImplementedError('Startup timeout type {} is not valid'.format(type_))
        return result


class BrokerManager(object):
    def __init__(self, process_holder: KafkaProcessHolder, kafka_dir: str, exhibitor: BukuExhibitor,
                 id_manager: BrokerIdGenerator, kafka_properties: KafkaProperties, timeout: StartupTimeout):
        self.kafka_dir = kafka_dir
        self.id_manager = id_manager
        self.exhibitor = exhibitor
        self.kafka_properties = kafka_properties
        self.process_holder = process_holder
        self.timeout = timeout

    def is_running_and_registered(self):
        if not self.process_holder.get():
            return False
        return self.id_manager.is_registered()

    def stop_kafka_process(self):
        self._terminate_process()
        self._wait_for_zk_absence()

    def _is_clean_election(self):
        value = self.kafka_properties.get_property('unclean.leader.election.enable')
        return value == 'false'

    def has_leadership(self):
        """
        Says if this broker is still a leader for partitions
        :return: True, if broker is a leader for some partitions.
        """
        broker_id = self.id_manager.get_broker_id()
        if not broker_id:
            return False
        return not self._is_leadership_transferred(dead_broker_ids=[broker_id])

    def _terminate_process(self):
        if self.process_holder.get() is not None:
            try:
                self.process_holder.get().terminate()
                self.process_holder.get().wait()
            except Exception as e:
                _LOG.error('Failed to wait for termination of kafka process', exc_info=e)
            finally:
                self.process_holder.set(None)

    def _wait_for_zk_absence(self):
        try:
            self.id_manager.wait_for_broker_id_absence()
        except Exception as e:
            _LOG.error('Failed to wait until broker id absence in zk', exc_info=e)

    def get_zk_connect_string(self):
        return self.kafka_properties.get_property('zookeeper.connect')

    def start_kafka_process(self, zookeeper_address):
        """
        Starts kafka using zookeeper address provided.
        :param zookeeper_address: Address to use for kafka
        :raise LeaderElectionInProgress: raised when broker can not be started because leader election is in progress
        """
        if not self.process_holder.get():
            if not self._is_leadership_transferred(active_broker_ids=self.exhibitor.get_broker_ids()):
                raise LeaderElectionInProgress()

            _LOG.info('Using ZK address: {}'.format(zookeeper_address))
            self.kafka_properties.set_property('zookeeper.connect', zookeeper_address)

            self.kafka_properties.dump()

            _LOG.info('Staring kafka process')
            self.process_holder.set(self._open_process())

            _LOG.info('Waiting for kafka to start up with timeout {} seconds'.format(self.timeout.get_timeout()))
            if not self.id_manager.wait_for_broker_id_presence(self.timeout.get_timeout()):
                self.timeout.on_timeout_fail()
                _LOG.error(
                    'Failed to wait for broker to start up, probably will kill, next timeout is'.format(
                        self.timeout.get_timeout()))

    def get_disjoined_isr_topic_partitions(self):
        """
        Gets list of (topic,partition) for which this broker should be in isr list, but for some reason it is not there 
        :return: list of tuples (topic:str, partition:int)
        """
        broker_id = int(self.id_manager.get_broker_id())
        checked_topic_partitoins = [
            (topic, partition) for topic, partition, broker_ids in self.exhibitor.load_partition_assignment()
            if broker_id in broker_ids
        ]
        unjoined_topic_partitions = []
        for topic, partition, state in self.exhibitor.load_partition_states():
            if (topic, partition) not in checked_topic_partitoins:
                continue
            if broker_id not in state.get('isr', []):
                unjoined_topic_partitions.append((topic, partition))
        return unjoined_topic_partitions

    def _is_leadership_transferred(self, active_broker_ids=None, dead_broker_ids=None):
        _LOG.info('Checking if leadership is transferred: active_broker_ids={}, dead_broker_ids={}'.format(
            active_broker_ids, dead_broker_ids))
        if self._is_clean_election():
            for topic, partition, state in self.exhibitor.load_partition_states():
                leader = str(state['leader'])
                if active_broker_ids and leader not in active_broker_ids:
                    if any(str(x) in active_broker_ids for x in state.get('isr', [])):
                        _LOG.warn(
                            'Leadership is not transferred for {} {} ({}, brokers: {})'.format(
                                topic, partition, json.dumps(state), active_broker_ids))
                        return False
                    else:
                        _LOG.warn('Shit happens! No single isr available for {}, {}, state: {}, '
                                  'skipping check for that'.format(topic, partition, json.dumps(state)))
                if dead_broker_ids and leader in dead_broker_ids:
                    _LOG.warn('Leadership is not transferred for {} {}, {} (dead list: {})'.format(
                        topic, partition, json.dumps(state), dead_broker_ids))
                    return False

        return True

    def _open_process(self):
        return subprocess.Popen(
            [self.kafka_dir + "/bin/kafka-server-start.sh", self.kafka_properties.settings_file])



import logging

import boto3
from instance_control import config
from instance_control import volume

from instance_control.aws import ec2_node
from instance_control.command import Command

_LOG = logging.getLogger('bubuku.cluster.command.attach')


class CreateCommand(Command):
    def __init__(self, cluster_config_path: str,
                 cluster_size: int,
                 availability_zone: str,
                 image_version: str):
        super().__init__(cluster_config_path)
        self.cluster_size = cluster_size
        self.availability_zone = availability_zone
        self.image_version = image_version

    def alter_config(self):
        if self.cluster_size:
            self.cluster_config['cluster_size'] = self.cluster_size
        if self.image_version:
            self.cluster_config['image_version'] = self.image_version
        self.cluster_config['create_ebs'] = True
        self.cluster_config['availability_zone'] = self.availability_zone

    def execute(self):
        ec2_node.create(self.cluster_config)
        ec2_client = boto3.client('ec2', region_name=self.cluster_config['region'])
        ec2_resource = boto3.resource('ec2', region_name=self.cluster_config['region'])
        volume.wait_volumes_attached(ec2_client, ec2_resource)
